package src.pellet

import collection.JavaConversions
import com.clarkparsia.owlapi.explanation.PelletExplanation
import com.clarkparsia.pellet.owlapiv3.PelletReasonerFactory
import de.derivo.sparqldlapi.QueryEngine
import java.io.{PrintStream, File}
import org.mindswap.pellet.PelletOptions
import org.mindswap.pellet.exceptions.InconsistentOntologyException
import org.mindswap.pellet.utils.progress.SilentProgressMonitor
import org.semanticweb.owlapi.apibinding.OWLManager
import org.semanticweb.owlapi.model._
import src.ship._

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 5.09.2012
 * Time: 09:44
 */

class PelletUpdateReasoner(ont: Ontology) extends JavaLogging {

  PelletExplanation.setup()
  PelletOptions.USE_UNIQUE_NAME_ASSUMPTION = true
  PelletOptions.USE_COMPLETION_QUEUE = true
  PelletOptions.USE_TRACING = true
  PelletOptions.USE_INCREMENTAL_CONSISTENCY = false
  PelletOptions.USE_SMART_RESTORE = false
  PelletOptions.USE_CACHING = true
  //PelletOptions.SATURATE_TABLEAU = true

  //println("SATURATE_TABLEAU "+PelletOptions.SATURATE_TABLEAU)

  val manager: OWLOntologyManager = OWLManager.createOWLOntologyManager()
  val pm = Builtin.prefixManager
  val owlont = Ontology2OWL.convert(ont, pm, manager)
  var reasoner = PelletReasonerFactory.getInstance().createNonBufferingReasoner(owlont)

  val progress = //new SwingProgressMonitor(2)
    new SilentProgressMonitor()

  reasoner.getKB.getTaxonomyBuilder.setProgressMonitor(progress)

  reasoner.getKB.setDoExplanation(true)
  manager.addOntologyChangeListener(reasoner)


  //println("Start prepare reasoner")
  reasoner.prepareReasoner()
  //println("Start realize KB0")
  try {
    reasoner.getKB.classify(); reasoner.getKB.realize()
    //println("Instances: "+reasoner.getKB.getIndividuals)
  }
  catch {
    case ex: InconsistentOntologyException =>
      log.error(ex, "Ontology:\n" + this)
  }

  //new uk.ac.manchester.cs.owlapi.modularity.SyntacticLocalityModuleExtractor()
  var engine = QueryEngine.create(manager, reasoner, true)

  def interrupt() {
    reasoner.interrupt()
  }

  def update(ont: Ontology): Pair[List[ABoxFormula], List[ABoxFormula]] = this.synchronized {

    val tempmanager: OWLOntologyManager = OWLManager.createOWLOntologyManager()
    val savedont = tempmanager.createOntology(owlont.getAxioms)

    val owlontnew = Ontology2OWL.convert(ont, pm, tempmanager)
    var changes = List.empty[OWLOntologyChange]
    for (old <- JavaConversions.asScalaSet(owlont.getAxioms)) {
      if (!owlontnew.containsAxiom(old)) {
        val change = new RemoveAxiom(owlont, old)
        //println("remove change: "+change)
        changes = changes ++ List(change)
      }
    }
    for (n <- JavaConversions.asScalaSet(owlontnew.getAxioms)) {

      if (!owlont.containsAxiom(n)) {
        val change = new AddAxiom(owlont, n)
        //println("add change: "+change)
        changes = changes ++ List(change)
      }
    }

    digestChanges(changes)

    var update: Pair[List[ABoxFormula], List[ABoxFormula]] = new Pair(Nil, Nil)

    val wasinconsistent = !reasoner.isConsistent
    if (wasinconsistent) {
      log.info("Reasoner inconsistent")
      val additionalChanges = restoreConsistency(savedont, owlontnew)
      changes = additionalChanges
      log.info("Additional " + additionalChanges)
      def toShipABox(change: OWLOntologyChange) = {
        val ax = change.getAxiom
        OWL2Ontology.convertABoxAxiom(ax).toList match {
          case hd :: Nil => hd
          case _ => throw new Exception("Cannot translate non ABox OWL Axiom: " + ax)
        }
      }
      val partition = additionalChanges filter
        (ax =>
          ax.getAxiom.isInstanceOf[OWLClassAssertionAxiom] ||
            ax.getAxiom.isInstanceOf[OWLObjectPropertyAssertionAxiom]
          ) partition (ax => ax.isInstanceOf[RemoveAxiom])
      update = Pair(partition._1 map toShipABox, partition._2 map toShipABox)
    } else log.info("Reasoner consistent")
    if (wasinconsistent) {
      /* WORKAROUND: Pelletincremental fails to answer completely roleassertions in some situations. Maybe a newer pelletreasoner
 * version will be better. Until then we simply, in the following, create a completely new reasoner
 * after having restored consistency */
      manager.removeOntologyChangeListener(reasoner)
      java.lang.System.gc()
      reasoner = PelletReasonerFactory.getInstance().createNonBufferingReasoner(owlont)
      reasoner.getKB.getTaxonomyBuilder.setProgressMonitor(progress)
      reasoner.getKB.setDoExplanation(true)
      manager.addOntologyChangeListener(reasoner)
      //println("Start prepare reasoner")
      reasoner.prepareReasoner()
      log.info("Start realize KB2")
      try {
        reasoner.getKB.classify(); reasoner.getKB.realize()
      }
      catch {
        case ex: InconsistentOntologyException => log.error(ex, "Ontology:\n" + this)
      }
      /* END WORKAROUND */
    }
    engine = QueryEngine.create(manager, reasoner, true)
    // log.info("Instances: " + reasoner.getKB.getIndividuals)
    update
  }

  private def power[A](s: Set[A]): List[Set[A]] = {
    //println("Power compute")
    @annotation.tailrec
    def pwr(s: List[A], acc: List[Set[A]]): List[Set[A]] = {
      s match {
        case Nil => acc
        case a :: as => pwr(as, acc ::: (acc map ((x: Set[A]) => x.+(a: A): Set[A])))
      }
    }
    pwr(s.toList, Set.empty[A] :: Nil).sortBy(_.size)
  }

  private def multiply[A](x: List[Set[A]], y: List[Set[A]]): List[Set[A]] = {
    log.info("Multiply compute")
    if (x.isEmpty) y
    else if (y.isEmpty) x
    else x flatMap (ex => y map (ey => ex ++ ey))
  }

  private def multiply[A](xs: Set[List[Set[A]]]): List[Set[A]] = {
    (xs :\ List.empty[Set[A]])((a, b) => multiply(a, b))
  }


  def restoreConsistency(owlontold: OWLOntology, owlontnew: OWLOntology): List[OWLOntologyChange] = this.synchronized {

    progress.setProgressTitle("Restoring consistency")
    progress.setProgress(5)
    val expGen: PelletExplanation = new PelletExplanation(reasoner)

    val explanations = JavaConversions.asScalaSet(expGen.getInconsistencyExplanations).toSet[java.util.Set[OWLAxiom]]
    log.info("Explanations = " + explanations)


    // Computing subsets

    var exps: Set[Set[OWLAxiom]] = Set.empty[Set[OWLAxiom]]
    for (s <- explanations) {
      exps = exps + JavaConversions.asScalaSet(s).toSet[OWLAxiom]
    }

    val temp = exps map (exp => {
      val inconsistentaxioms = exp filter
        (ax => ax.isInstanceOf[OWLClassAssertionAxiom] || ax.isInstanceOf[OWLObjectPropertyAssertionAxiom])
      //println("Assertions only: "+inconsistentaxioms)
      val keepaxioms = inconsistentaxioms filter (ax => owlontnew.containsAxiom(ax) && !owlontold.containsAxiom(ax))
      //println("New assertions only"+keepaxioms)
      val removeaxioms: Set[OWLAxiom] = inconsistentaxioms.--(keepaxioms)
      //println("Remove axioms: "+removeaxioms)
      power[OWLAxiom](removeaxioms)
    })
    log.info("Powersets done")
    //println("temp = "+temp)

    var outercounter: Int = 0
    var changes = List.empty[OWLOntologyChange]

    while (changes.isEmpty && outercounter <= temp.size) {
      outercounter = outercounter + 1
      val subsets: List[Set[OWLAxiom]] = multiply(temp.take(outercounter))
      log.info("Multiply " + outercounter + " done")
      //println("remove axiom subsets = "+subsets)

      val it = subsets.iterator

      var counter = 0

      while (changes.isEmpty && it.hasNext) {
        counter = counter + 1

        progress.setProgress((counter + 100) / subsets.length)

        val axsubst = it.next()
        for (ax <- axsubst) {
          //log.info("Removing "+OWL2Ontology.OWLAxiom2abox(ax))
          changes = new RemoveAxiom(owlont, ax) :: changes
        }

        digestChanges(changes)

        log.info("Consistency = " + reasoner.isConsistent)
        if (!reasoner.isConsistent) {
          changes = List.empty[OWLOntologyChange]
          for (ax <- axsubst) {
            //println("Adding "+OWL2Ontology.OWLAxiom2abox(ax))
            changes = new AddAxiom(owlont, ax) :: changes
          }
          digestChanges(changes)
          changes = List.empty[OWLOntologyChange]
        }
      }
    }
    progress.setProgress(100)
    progress.taskFinished()
    changes
  }


  private def digestChanges(changes: List[OWLOntologyChange]) {
    //println("Applying changes "+changes)
    this.synchronized {
      //reasoner.processChanges(JavaConversions.asJavaList(changes))
      manager.applyChanges(JavaConversions.seqAsJavaList(changes))
      /* for (owlontchange <- changes) {
      if (owlontchange.isInstanceOf[AddAxiom]) {
         val addaxiom = owlontchange.asInstanceOf[AddAxiom]
         manager.addAxiom(owlont,addaxiom.getAxiom)
       } else if (owlontchange.isInstanceOf[RemoveAxiom]) {
         val removeaxiom = owlontchange.asInstanceOf[RemoveAxiom]
         manager.removeAxiom(owlont,removeaxiom.getAxiom)
       }
     } */

      this.synchronized {
        reasoner.refresh() // Apparently necessary to delete old cache entries from queries
        reasoner.getKB.setDoExplanation(true)
        reasoner.prepareReasoner()
        log.info("Start realize KB1")
        try {
          reasoner.getKB.classify()
          reasoner.getKB.realize()
        }
        catch {
          case e: InconsistentOntologyException =>
            log.info("inconsistent ontology exception")
            //e.printStackTrace()
        }
      }
      log.info("Reasoner consistency = " + reasoner.getKB.isConsistent)
      //reasoner.refresh()
      //reasoner.flush()
      //reasoner.prepareReasoner()
    }

  }


}

