package src.ship

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 15.06.2012
 * Time: 10:54
 */

sealed trait Effect extends ObjectLevelVariable {
  def subst(s: Substitution): Effect
}

case class UnconditionalEffect(newaboxformulas: List[ABoxFormula], delete: List[VarOrConstant]) extends Effect {
  override def toString = newaboxformulas.toString + (if (delete.isEmpty) "" else "delete(" + delete + ")")

  def subst(s: Substitution): UnconditionalEffect = {
    UnconditionalEffect(newaboxformulas map s.apply, delete map s.apply)
  }

  override def freeVars: Set[ObjectLevelVariable] =
    newaboxformulas.map(_.freeVars).fold(Set.empty)(_ union _) union (delete collect {
      case x: Var => x
    }).toSet

  def merge(other: UnconditionalEffect): UnconditionalEffect = {
    UnconditionalEffect((newaboxformulas ++ other.newaboxformulas).distinct,
      (delete ++ other.delete).distinct)
  }
}

case class ConditionalEffect(condeffects: Map[List[ABoxFormula], UnconditionalEffect]) extends Effect {
  override def toString = (condeffects.keys :\ "")((key, rest) => "if (" + key + ") " + condeffects.get(key).get)

  def subst(s: Substitution): ConditionalEffect = {
    ConditionalEffect(
      (condeffects.keys :\ Map[List[ABoxFormula], UnconditionalEffect]())
        ((phis, m2: Map[List[ABoxFormula], UnconditionalEffect]) =>

          Map[List[ABoxFormula], UnconditionalEffect]((phis map s.apply) -> (condeffects get phis).get.subst(s))
            ++ m2))
  }

  override def freeVars: Set[ObjectLevelVariable] = {
    val temp1 = (condeffects.keySet :\ List.empty[ABoxFormula])(_ ++ _)
    // The variables in the consition and effects formulas
    (temp1 :\ Set.empty[ObjectLevelVariable])((l, s) => l.freeVars ++ s) ++
      // The variables in the effect/delete parts.
      (condeffects.keySet :\ Set.empty[ObjectLevelVariable])((k, s) => condeffects.get(k).get.freeVars ++ s)
  }
}

object MkString {
  def commaSep(l: List[Any], start: String, end: String) = l.mkString(start, ", ", end)

  def paramList(name: String, l: List[Any]) = if (l.isEmpty) name else commaSep(l, name + "(", ")")
}

sealed abstract class DLDeclaration extends JavaLogging {
  def make_subst(fpars: List[Var], apars: List[VarOrConstant]): Substitution = {
    assert(fpars.length == apars.length)
    Pair(fpars, apars) match {
      case Pair(Var(n) :: xs, y :: ys) => CompSubst(/(y, Var(n)), make_subst(xs, ys))
      case Pair(Nil, Nil) => Id()
    }
  }
}

case class IndirectEffect(name: String, init: List[ABoxFormula], effect: UnconditionalEffect, cond: List[ABoxFormula])
  extends DLDeclaration {
  override def toString = MkString.commaSep(init,
    "indirect effect " + name + " = {\n init = ",
    "\n causes = " + effect) +
    MkString.commaSep(cond, "\n cond = ", "\n}")

  def apply(o: Ontology, currenteffects: List[ABoxFormula]): Option[List[ABoxFormula]] = {
    val subont = Ontology(o.tbox, o.rbox, currenteffects)
    subont.query(init) match {
      case Nil => None
      case l@_ :: _ =>
        var res: Option[List[ABoxFormula]] = None
        val it = l.iterator
        while (it.hasNext && res == None) {
          res = testinstance(o, subont, currenteffects, it.next())
        }
        res
    }
  }

  def testinstance(o: Ontology, subont: Ontology, currenteffects: List[ABoxFormula], s: Substitution): Option[List[ABoxFormula]] = {
    log.info("Testing instance " + s)
    val initinstance = s.apply(init)
    o.query(initinstance) match {
      case s1 :: _ => // The init causes are already implied, skip
        log.info("Init already implied " + initinstance)
        None
      case Nil => // Ok, the init causes are not already implied before
        var found = false
        var subst: Substitution = s
        if (initinstance forall (phiraw => {
          val phi = subst.apply(phiraw)
          subont.query(phi) match {
            case Nil => o.query(phi) match {
              case Nil => false
              case s1 :: _ => if (!subont.mycontains(OntologyFun.Neg(s1.apply(phi)), currenteffects)) {
                subst = CompSubst(s1, subst)
                true
              } else false
            }
            case s1 :: _ => subst = CompSubst(s1, subst); found = true; true
          }
        })) {
          // we have all inits proved, and subst is the substitution
          // Still test if at least one is new and if the conditions are all true in the old ontolgoy
          if (found) {
            o.query(subst.apply(cond)) match {
              case slist@_ :: _ => // yes, the conditions are fulfilled, adding the derived effects
                slist flatMap
                  (s => {
                    val xsubst = CompSubst(s, subst)
                    if (subont.query(xsubst.apply(effect.newaboxformulas)).isEmpty) {
                      // The indirect effects are not already contained
                      xsubst.apply(effect.newaboxformulas)
                    } else List.empty[ABoxFormula]
                  }) match {
                  case Nil => None // Nothing was added
                  case v => Some(currenteffects ++ v)
                }
              case Nil => // Conditions not fulfilled
                None
            }
          } else {
            // nothing was new, nothing to add
            None
          }
        } else {
          // Some inits  failed to be proved
          None
        }
    }
  }
}

case class Action(name: String, formalpars: List[Var],
                  pre: List[ABoxFormula],
                  exec: Option[Pair[java.lang.reflect.Method, List[Var]]],
                  effect: UnconditionalEffect,
                  condeffects: ConditionalEffect) extends DLDeclaration {
  override def toString = MkString.paramList(name, formalpars)

  private def generateNewIndividuals(o: Ontology, finalsubst: Substitution): Substitution = {
    var result = finalsubst
    for (fv <- effect.subst(finalsubst).freeVars) {
      result = CompSubst(result, /(Const(o.newIndividualGenerator), fv.asInstanceOf[Var]))
    }
    result
  }

  def apply(o: Ontology, actpars: List[VarOrConstant]): Option[Ontology] = {

    val subst = make_subst(formalpars, actpars)
    val preinst: List[ABoxFormula] = pre map (phi => subst.apply(phi))

    val substitutions = o.query(preinst)
    if (!substitutions.isEmpty) {
      //println("substs "+substitutions)
      var finalsubst: Substitution = CompSubst(substitutions.head, subst)
      //println("Matched "+preinst)
      //println("subst"+finalsubst)
      //println((effect.subst(finalsubst)))

      finalsubst = generateNewIndividuals(o, finalsubst)
      //println(finalsubst)

      assert(effect.subst(finalsubst).freeVars.isEmpty)

      var execeffect = UnconditionalEffect(Nil, Nil)

      exec match {
        case Some(Pair(m, pars)) =>
          invokeMethod(o, m, pars map (v => finalsubst.apply(v).name)) match {
            case Some(ue) => execeffect = ue
            case v => log.warning("program result empty")
          }
        case v => log.info("no external exec for action: " + name)
      }

      var finaleffect = effect.subst(finalsubst).merge(execeffect)
      val it = condeffects.condeffects.keysIterator
      while (it.hasNext) {
        val key = it.next()
        val condsubsts = o.query(key map (phi => finalsubst.apply(phi)))
        if (!condsubsts.isEmpty) {
          finalsubst = CompSubst(condsubsts.head, finalsubst)
          val effect = condeffects.condeffects.get(key).get.subst(finalsubst)
          assert(effect.freeVars.isEmpty)
          finaleffect = finaleffect.merge(effect)
        }
      }
      val newont = o.updatesimple(finaleffect.newaboxformulas, finaleffect.delete)
      newont
    } else {
      log.info("No Match for '" + name + "' " + preinst)
      None
    }
  }

  def invokeMethod(o: Ontology, m: java.lang.reflect.Method, args: List[String]): Option[UnconditionalEffect] = {
    //val clazz = Class.forName("src.ship.external.PathCoordination$"); // Note the trailing '$'
    //println(clazz)
    //val field = clazz.getField("MODULE$");
    //val instance = field.get(null);

    /*val a = args.toArray[String]
    val res = args.length match {
      case 0  => m.invoke(null,o)
      case 1  => m.invoke(null,o,a.apply(0))
      case 2  => m.invoke(null,o,a.apply(0),a.apply(1))
      case 3  => m.invoke(null,o,a.apply(0),a.apply(1),a.apply(2))
      case 4  => m.invoke(null,o,a.apply(0),a.apply(1),a.apply(2),a.apply(3))
      case 5  => m.invoke(null,o,a.apply(0),a.apply(1),a.apply(2),a.apply(3),a.apply(4))
      case 6  => m.invoke(null,o,a.apply(0),a.apply(1),a.apply(2),a.apply(3),a.apply(4),a.apply(5))
      case 7  => m.invoke(null,o,a.apply(0),a.apply(1),a.apply(2),a.apply(3),a.apply(4),a.apply(5),a.apply(6))
      case 8  => m.invoke(null,o,a.apply(0),a.apply(1),a.apply(2),a.apply(3),a.apply(4),a.apply(5),a.apply(6),a.apply(7))
      case 9  => m.invoke(null,o,a.apply(0),a.apply(1),a.apply(2),a.apply(3),a.apply(4),a.apply(5),a.apply(6),a.apply(7),a.apply(8))
      case 10 => m.invoke(null,o,a.apply(0),a.apply(1),a.apply(2),a.apply(3),a.apply(4),a.apply(5),a.apply(6),a.apply(7),a.apply(8),a.apply(9))
      case _ => None
    }*/
    m.invoke(null, o :: args: _*) match {
      case x: Option[_] => x match {
        case Some(y) => y match {
          case effect: UnconditionalEffect => Some(effect)
          case _ => None
        }
        case None => None
      }
      case _ => None
    }
  }
}

case class Function(name: String, formalpars: List[Var], body: DLProgram) extends DLDeclaration {
  def expand(actpars: List[VarOrConstant]): DLProgram = {
    var subst: Substitution = null
    try {
      subst = make_subst(formalpars, actpars)
    }
    catch {
      case e: AssertionError => log.error(e, "Failure expanding " + name)
    }
    //println("Expansion subst "+subst)
    body.subst(subst)
  }
}

case class Predicate(name: String, formalpars: List[Var], desc: String, body: FOLTL[Concept, ABoxFormula]) extends DLDeclaration {
  def expand(actpars: List[VarOrConstant]): DLProgram = {
    val subst = make_subst(formalpars, actpars)
    //println("Expanding "+name+" with desc '"+desc+"'")
    MonitorExec(subst.applyF(DefPred(Pred(name, formalpars))), subst.applyS(desc), subst.applyF(body))
  }
}

sealed abstract class DLProgStepResult {
  var stutteringmonitors: List[MonitorExec] = List.empty[MonitorExec]

  def extendBy(other: DLProgStepResult): DLProgStepResult = {
    stutteringmonitors = other.getStutteringMonitors ++ stutteringmonitors
    this
  }

  def getStutteringMonitors: List[MonitorExec] = {
    stutteringmonitors
  }
}

case class Done(o: Ontology, p: DLProgram) extends DLProgStepResult

case class Step(o: Ontology, p: DLProgram, cont: DLEvaluation) extends DLProgStepResult

case class Fail(o: Ontology, p: DLProgram) extends DLProgStepResult

case class Stutter(o: Ontology, p: DLProgram, cont: DLEvaluation) extends DLProgStepResult {
  override def toString: String = this.getClass.getSimpleName + "(" + p + "," + cont + ")"

  override def getStutteringMonitors: List[MonitorExec] = {
    p match {
      case mon@MonitorExec(_, _, _) => mon :: stutteringmonitors
      case _ => stutteringmonitors
    }
  }
}

case class Error(msg: String) extends DLProgStepResult

case class DLEvaluation(decls: Map[String, DLDeclaration], p: DLProgram) {
  override def toString: String = this.getClass.getSimpleName + "(continuation " + p + ")"

  def getDescription(n: String, actpars: List[VarOrConstant]): String = {
    decls get n match {
      case Some(Predicate(name, fpars, desc, b)) => {
        Pred(n, actpars).expand(Predicate(name, fpars, desc, b)) match {
          case MonitorExec(phi, idesc, psi) => idesc
          case _ => ""
        }
      }
      case _ => ""
    }
  }

  def eval_step(o: Ontology): DLProgStepResult = {
    p match {
      case Skip => Done(o, Skip)
      case Appl(n, actpars) => decls get n match {
        case Some(Action(_, _, _, _, _, _)) => DLEvaluation(decls, Act(n, actpars)).eval_step(o)
        case Some(Function(_, _, _)) => DLEvaluation(decls, Fct(n, actpars)).eval_step(o)
        case _ => Error("Undeclared action of function " + n)
      }
      case Act(n, actpars) => decls get n match {
        case Some(a@Action(_, _, _, _, _, _)) => Act(n, actpars).apply(o, a) match {
          case Some(newont) => Done(newont, Act(n, actpars))
          case None => Stutter(o, Act(n, actpars), this)
        }
        case _ => Error("Undeclared action " + n)
      }
      case Fct(n, actpars) => decls get n match {
        case Some(a@Function(_, _, _)) => DLEvaluation(decls, Fct(n, actpars).expand(a)).eval_step(o)
        case _ => Error("Undeclared function " + n)
      }
      case Pred(n, actpars) => decls get n match {
        case Some(a@Predicate(_, _, _, _)) => DLEvaluation(decls, Pred(n, actpars).expand(a)).eval_step(o)
        case _ => Error("Undeclared predicate " + n)
      }
      case mon@MonitorExec(orig, desc, phi) => {
        //println("Here "+desc+" "+p)
        val newmon = mon.lastphi match {
          case None =>
            val progphi = FOLTL.progress(o, phi, decls)
            val newdesc = orig match {
              case v@DefPred(Pred(n, actpars)) => getDescription(n, actpars)
              case _ => desc
            }
            val newmon = MonitorExec(orig, newdesc, progphi)
            newmon.lastphi = Some(phi)
            newmon
          case Some(_) => mon
        }
        newmon.phi match {
          case Not(True) => Fail(o, newmon)
          case True => Done(o, newmon)
          case _ => Stutter(o, newmon, DLEvaluation(decls, newmon))
          //Stutter(o,MonitorExec(orig, phi),DLEvaluation(decls,MonitorExec(orig,newphi)))
          //else Step(o,MonitorExec(orig,newphi),DLEvaluation(decls,MonitorExec(orig,newphi)))
        }
      }
      case Seq(p1, p2) => DLEvaluation(decls, p1).eval_step(o) match {
        case Done(newont, step) =>
          /* if (step.isInstanceOf[MonitorExec])
            DLEvaluation(decls,p2).eval_step(o)
          else */ Step(newont, step, DLEvaluation(decls, p2))
        case Fail(newont, step) => Fail(newont, step)
        case Step(newont, step, DLEvaluation(decls1, p1rest)) => Step(newont, step, DLEvaluation(decls1, Seq(p1rest, p2)))
        case Stutter(newont, step, DLEvaluation(decls1, p1rest)) => Stutter(newont, step, DLEvaluation(decls1, Seq(p1rest, p2)))
        case Error(msg) => Error(msg)
      }
      case TestNot(phi) => if (TestNot(phi).eval(o)) {
        Done(o, TestNot(phi))
      } else if (Test(phi).eval(o)) {
        Fail(o, TestNot(phi))
      } else {
        Stutter(o, TestNot(phi), DLEvaluation(decls, TestNot(phi)))
      }
      case Test(phi) => if (Test(phi).eval(o)) {
        Done(o, Test(phi))
      } else if (TestNot(phi).eval(o)) {
        Fail(o, Test(phi))
      } else {
        Stutter(o, Test(phi), DLEvaluation(decls, Test(phi)))
      }
      case Star(p1) => DLEvaluation(decls, p1).eval_step(o) match {
        case Done(newont, step) =>
          /* if (step.isInstanceOf[MonitorExec])
            DLEvaluation(decls,Star(p1)).eval_step(o)
          else */ Step(newont, step, DLEvaluation(decls, Star(p1)))
        case Fail(newont, step) => Done(newont, step)
        case Step(newont, step, DLEvaluation(decls1, p1rest)) => Step(newont, step, DLEvaluation(decls1, Seq(p1rest, Star(p1))))
        case Stutter(newont, step, DLEvaluation(decls1, p1rest)) => Stutter(newont, step, DLEvaluation(decls1, Star(p1)))
        case Error(msg) => Error(msg)
      }
      case Alt(p1, p2) => DLEvaluation(decls, p1).eval_step(o) match {
        case Done(newont, step) => Done(newont, step)
        case Fail(newont, step) => Step(newont, step, DLEvaluation(decls, p2))
        case Step(newont, step, DLEvaluation(decls1, p1rest)) => Step(newont, step, DLEvaluation(decls1, Alt(p1rest, p2)))
        case Stutter(newont, step, DLEvaluation(decls1, p1rest)) => Stutter(newont, step, DLEvaluation(decls1, Alt(p1, p2)))
        case Error(msg) => Error(msg)
      }
      case Guard(phis, prg) => throw new Exception("eval_step unexpected guard: " + p)
      case Switch(cases) =>
        val it = cases.iterator
        var program: Option[DLProgram] = None
        while (it.hasNext && program == None) {
          it.next() match {
            case Guard(phis, prg) => o.query(phis) match {
              case subst :: substs => program = Some(prg.subst(subst))
              case v => v
            }
            case v => v
          }
        }
        program match {
          case Some(prg) => DLEvaluation(decls, prg).eval_step(o)
          case None => Stutter(o, p, DLEvaluation(decls, p))
        }
      case Forall(phis, prg) =>
        val programm = o.query(phis) match {
          case subst :: substs => (prg.subst(subst) /: substs)((p, s) => Par(p, prg.subst(s)))
          case Nil => Skip
        }
        DLEvaluation(decls, programm).eval_step(o)
      case Par(p1, p2) => DLEvaluation(decls, p1).eval_step(o) match {
        case x@Done(newont, step) =>
          if (step.isInstanceOf[MonitorExec])
            DLEvaluation(decls, p2).eval_step(o).extendBy(x)
          else Step(newont, step, DLEvaluation(decls, p2)).extendBy(x)
        case x@Fail(_, _) => x
        case x@Step(newont, step, DLEvaluation(decls1, p1rest)) =>
          Step(newont, step, DLEvaluation(decls1, Par(p2, p1rest))).extendBy(x)
        case Error(msg) => Error(msg)
        case s1@Stutter(newont, step, DLEvaluation(decls1, p1rest)) =>
          val s2 = DLEvaluation(decls1, p2).eval_step(newont)
          s2 match {
            case Stutter(newont2, step2, DLEvaluation(decls2, p2rest)) =>
              Stutter(newont2, step2, DLEvaluation(decls2, Par(p2rest, p1rest))).extendBy(s1).extendBy(s2)
            case Done(newont2, step2) => Step(newont2, step2, DLEvaluation(decls1, p1rest)).extendBy(s1).extendBy(s2)
            case Step(newont2, step2, DLEvaluation(decls2, p2rest)) =>
              Step(newont2, step2, DLEvaluation(decls2, Par(p1rest, p2rest))).extendBy(s1).extendBy(s2)
            case Fail(_, _) => s2.extendBy(s1)
            case Error(_) => s2.extendBy(s1)
          }
      }
    }
  }
}


sealed abstract class DLProgram extends ObjectLevelVariable with GroundableObject[DLProgram] {
  def subst(s: Substitution): DLProgram

  def groundActPars(exceptions: Set[ObjectLevelVariable], actpars: List[VarOrConstant]): List[VarOrConstant] =
    actpars map {
      case v@Var(n) => if (exceptions.contains(v)) v else Const(n)
      case e => e
    }

  def freeActParsVars(actpars: List[VarOrConstant]): Set[ObjectLevelVariable] =
    (actpars collect {
      case v: Var => v
    }).toSet

  def varsOfAboxes(aBoxes: List[ABoxFormula]): Set[ObjectLevelVariable] =
    aBoxes.map(_.freeVars).fold(Set.empty)(_ union _)

  def substActPars(s: Substitution, actpars: List[VarOrConstant]): List[VarOrConstant] =
    actpars map s.apply
}

case class Appl(name: String, actpars: List[VarOrConstant]) extends DLProgram {
  def subst(s: Substitution): DLProgram =
    Appl(name, substActPars(s, actpars))

  def groundBy(exceptions: Set[ObjectLevelVariable]) =
    Appl(name, groundActPars(exceptions, actpars))

  override def toString = MkString.paramList(name, actpars)

  override def freeVars: Set[ObjectLevelVariable] = freeActParsVars(actpars)
}

case object Skip extends DLProgram {
  override def toString = "skip"

  def subst(s: Substitution): DLProgram = Skip

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = Skip

  override def freeVars: Set[ObjectLevelVariable] = Set.empty[ObjectLevelVariable]
}

case class Act(name: String, actpars: List[VarOrConstant]) extends DLProgram {
  override def toString = MkString.paramList(name, actpars)

  def subst(s: Substitution): DLProgram =
    Act(name, substActPars(s, actpars))

  def apply(o: Ontology, decl: Action): Option[Ontology] = {
    assert(name == decl.name)
    decl.apply(o, actpars)
  }

  override def groundBy(exceptions: Set[ObjectLevelVariable]) =
    Act(name, groundActPars(exceptions, actpars))

  override def freeVars: Set[ObjectLevelVariable] = freeActParsVars(actpars)
}

case class Fct(name: String, actpars: List[VarOrConstant]) extends DLProgram {
  override def toString = MkString.paramList(name, actpars)

  def subst(s: Substitution): DLProgram =
    Fct(name, substActPars(s, actpars))

  def expand(fdecl: Function): DLProgram = {
    assert(name == fdecl.name)
    val body = fdecl.expand(actpars)
    //println("Expanded "+Fct(name, actpars)+" to "+body)
    body
  }

  override def groundBy(exceptions: Set[ObjectLevelVariable]) =
    Fct(name, groundActPars(exceptions, actpars))

  override def freeVars: Set[ObjectLevelVariable] = freeActParsVars(actpars)
}

case class Pred(name: String, actpars: List[VarOrConstant]) extends DLProgram {
  override def toString = "init " + MkString.paramList(name, actpars)

  def subst(s: Substitution): DLProgram =
    Pred(name, substActPars(s, actpars))

  def expand(pdecl: Predicate): DLProgram = {
    assert(name == pdecl.name)
    val body = pdecl.expand(actpars)
    //println("Expanded "+Pred(name, actpars)+" to "+body)
    body
  }

  override def groundBy(exceptions: Set[ObjectLevelVariable]) =
    Pred(name, groundActPars(exceptions, actpars))

  override def freeVars: Set[ObjectLevelVariable] = freeActParsVars(actpars)
}

case class MonitorExec(orig: FOLTL[Concept, ABoxFormula], desc: String, phi: FOLTL[Concept, ABoxFormula]) extends DLProgram {
  var lastphi: Option[FOLTL[Concept, ABoxFormula]] = None

  override def toString = "[" + desc + ": " + phi.toString + "]"

  override def subst(s: Substitution): DLProgram = {
    MonitorExec(s.applyF(orig), s.applyS(desc), s.applyF(phi))
  }

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = {
    val varsubst = (orig.freeVars -- exceptions :\ (Id(): Substitution))((v, subst) =>
      CompSubst(/(Const(v.asInstanceOf[Var].name), v.asInstanceOf[Var]), subst))

    MonitorExec(orig.groundBy(exceptions), varsubst.applyS(desc), phi.groundBy(exceptions))
  }

  override def freeVars: Set[ObjectLevelVariable] = orig.freeVars union phi.freeVars
}

case class Test(phi: List[ABoxFormula]) extends DLProgram {
  override def toString = phi.toString + "?"

  def subst(s: Substitution): DLProgram =
    Test(phi map s.apply)

  override def groundBy(exceptions: Set[ObjectLevelVariable]) =
    Test(phi map (_.groundBy(exceptions)))

  def eval(o: Ontology): Boolean = {
    val res = !o.query(phi).isEmpty
    //println("test sat "+phi+" = "+res)
    res
  }

  override def freeVars: Set[ObjectLevelVariable] =
    varsOfAboxes(phi)
}

case class TestNot(phi: List[ABoxFormula]) extends DLProgram {
  override def toString = "not " + phi.toString + "?"

  def subst(s: Substitution): DLProgram =
    TestNot(phi map s.apply)

  override def groundBy(exceptions: Set[ObjectLevelVariable]) =
    TestNot(phi map (_.groundBy(exceptions)))

  def eval(o: Ontology): Boolean = {
    val res = o.isUnsatisfiable(phi)
    //println("test unsat "+phi+" = "+res)
    res
  }

  override def freeVars: Set[ObjectLevelVariable] =
    varsOfAboxes(phi)
}

case class Alt(p1: DLProgram, p2: DLProgram) extends DLProgram {
  override def toString = p1.toString + "+>" + p2

  def subst(s: Substitution): DLProgram =
    Alt(p1.subst(s), p2.subst(s))

  override def groundBy(exceptions: Set[ObjectLevelVariable]) =
    Alt(p1.groundBy(exceptions), p2.groundBy(exceptions))

  override def freeVars: Set[ObjectLevelVariable] =
    p1.freeVars union p2.freeVars
}

case class Guard(phi: List[ABoxFormula], p: DLProgram) extends DLProgram {
  override def toString = "case " + phi + " => " + p

  def subst(s: Substitution): Guard =
    Guard(phi map s.apply, p.subst(s))

  override def groundBy(exceptions: Set[ObjectLevelVariable]) =
    Guard(phi, p.groundBy(exceptions ++ varsOfAboxes(phi)))

  override def freeVars: Set[ObjectLevelVariable] = {
    p.freeVars -- varsOfAboxes(phi)
  }
}

case class Forall(phi: List[ABoxFormula], p: DLProgram) extends DLProgram {
  override def toString = "forall " + phi + " => " + p

  def subst(s: Substitution): Forall =
    Forall(phi map s.apply, p.subst(s))

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = {
    Forall(phi, p.groundBy(exceptions ++ varsOfAboxes(phi)))
  }

  override def freeVars: Set[ObjectLevelVariable] = {
    p.freeVars -- varsOfAboxes(phi)
  }
}

case class Switch(cs: List[Guard]) extends DLProgram {
  override def toString = "switch " + cs

  override def subst(s: Substitution): DLProgram =
    Switch(cs map (a => a.subst(s)))

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = {
    Switch(cs map (f => f.groundBy(exceptions)))
  }

  override def freeVars: Set[ObjectLevelVariable] = {
    var acc = Set.empty[ObjectLevelVariable]
    for (par <- cs) acc = acc union par.freeVars
    acc
  }
}

case class Par(p1: DLProgram, p2: DLProgram) extends DLProgram {
  override def toString = p1.toString + "|" + p2

  override def subst(s: Substitution): DLProgram =
    Par(p1.subst(s), p2.subst(s))

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = {
    Par(p1.groundBy(exceptions), p2.groundBy(exceptions))
  }

  override def freeVars: Set[ObjectLevelVariable] = {
    p1.freeVars union p2.freeVars
  }
}

case class Seq(p1: DLProgram, p2: DLProgram) extends DLProgram {
  override def toString = p1.toString + ";" + p2

  def subst(s: Substitution): DLProgram =
    Seq(p1.subst(s), p2.subst(s))

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = {
    Seq(p1.groundBy(exceptions), p2.groundBy(exceptions))
  }

  override def freeVars: Set[ObjectLevelVariable] = {
    p1.freeVars union p2.freeVars
  }
}

case class Star(p: DLProgram) extends DLProgram {
  override def toString = p.toString + "*"

  def subst(s: Substitution): DLProgram =
    Star(p.subst(s))

  override def groundBy(exceptions: Set[ObjectLevelVariable]) = {
    Star(p.groundBy(exceptions))
  }

  override def freeVars: Set[ObjectLevelVariable] = {
    p.freeVars
  }
}

/*
trait DLProgramParser extends ProgramAndMonitorParser

object DLProgram extends DLProgramParser {
  val assign = Action("assignWC", List(Var("P"), Var("WC"), Var("L")),
    // Preconditions
    List(
      Ontology.RV("TravelRequest", "P", "L"),
      Ontology.CV("Person", "P"),
      Ontology.CV("Location", "L"),
      Ontology.CV("WheelChair", "WC"),
      Ontology.CV("Free", "WC")//,
      //RoleAssertion(Var("P"),Var("L"),NotRole(NamedRole("At")))
    ),
  // Effects
    UnconditionalEffect(List(Ontology.RV("Assigned","WC","P"),RoleAssertion(Var("WC"),Var("P"),NotRole(NamedRole("Free"))))),
  // Conditional Effects
  ConditionalEffect(Map.empty))

  val unassign = Action("unassignWC", List(Var("P"), Var("WC"), Var("L")),
    // Preconditions
    List(
      Ontology.RV("TravelRequest", "P", "L"),
      Ontology.CV("Person", "P"),
      Ontology.CV("Location", "L"),
      Ontology.CV("WheelChair", "WC"),
      Ontology.RV("Assigned","WC","P"),
      RoleAssertion(Var("P"),Var("L"),NamedRole("At"))
    ),
    // Effects
    UnconditionalEffect(List(Ontology.RV("Free","WC","P"),
      RoleAssertion(Var("P"),Var("L"),NotRole(NamedRole("TravelRequest"))),
        RoleAssertion(Var("WC"),Var("P"),NotRole(NamedRole("Assigned"))))),
    // Conditional Effects
    ConditionalEffect(Map.empty))

  val assignunassign =
    Function("AU",List(Var("P"),Var("WC"),Var("L")),
             Seq(Act("assignWC",List(Var("P"),Var("WC"),Var("L"))),Act("unassignWC",List(Var("P"),Var("WC"),Var("L")))))
  var decls = Map("assignWC"->assign,"unassignWC"->unassign,"AU"->assignunassign)

  var prog = Fct("AU",List(Const("adam"),Const("Rolland"),Const("bedroom")))

  def dle = DLEvaluation(decls,prog)

  dle.eval_step(Ontology.o)

  var stringdecls =
    "action Assign(p) = { pre = w:WheelChair, p:Person, w:Free, (p,r):TravelGoal, p:Unassigned " +
    "     effect = (w,p):Assigned }" +
    "action Unassign(p) = { pre = w:WheelChair, p:Person, w:~Free, (p,r):TravelGoal, (w,p):Assigned, (p,r):At " +
    "     effect = (w,p):~ Assigned }" +
    "action Skip = { pre = }" +
    "function AUsimple(p) = { " +
    "    if (p:person and (p,r):TravelGoal and p:Unassigned) " +
    "        Assign(p);Unassign(p) " +
    "    else skip }"

}
      */
