package src.ship

import com.clarkparsia.owlapi.explanation.PelletExplanation
import java.io._
import org.coode.owlapi.owlxml.renderer.OWLXMLRenderer
import org.semanticweb.owlapi.apibinding.OWLManager
import org.semanticweb.owlapi.model.{OWLObject, OWLAnnotation, OWLOntologyManager}
import org.semanticweb.owlapi.reasoner.OWLReasoner
import scala.collection.JavaConversions
import scala.collection.immutable.Nil
import scala.util.parsing.combinator.JavaTokenParsers
import src.pellet.PelletUpdateReasoner

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 13.06.2012
 * Time: 17:04
 */

trait OntologyParser extends JavaTokenParsers
with TBoxFormulaParser
with RBoxFormulaParser
with ABoxFormulaParser
with OntMorphismParser
with JavaLogging {
  protected override val whiteSpace = """(\s|//.*|(?m)/\*(\*(?!/)|[^*])*\*/)+""".r

  def ontology: Parser[Ontology] = ontology(System.getProperty("user.dir") + File.separator + "examples")

  def ontology(currentpath: String): Parser[Ontology] = {
    opt(imports(currentpath)) ~ opt("%TBOX" ~ rep(tboxformula)) ~ opt("%RBOX" ~ rep(rboxformulas)) ~ rep(aboxformula) ^^ {
      case ont ~ tboxopt ~ rboxopt ~ aboxphis =>

        val importont = ont match {
          case Some(o) => o
          case None => EmptyOntology
        }
        val tboxont = tboxopt match {
          case Some(_ ~ tboxphis) => (importont /: tboxphis)((o, o1) => o.extendBy(o1))
          case None => importont
        }
        val rbox = rboxopt match {
          case Some(_ ~ rboxphis) => rboxphis flatMap (x => x)
          case None => List.empty[RBoxFormula]
        }
        val abox = aboxphis map (_.ground)


        val res = tboxont.extendBy(Ontology(Nil, rbox, abox))
        //println("Parsed Ontology "+res.tbox.size+"/"+res.rbox.size+"/"+res.abox.size)
        res
    }
  }

  def imports(currentpath: String): Parser[Ontology] =
    rep("import" ~ pathparser ~ opt("with" ~ ontmorphparser)) ^^ {
      case imports =>
        var tbox = List.empty[TBoxFormula]
        var rbox = List.empty[RBoxFormula]
        var abox = List.empty[ABoxFormula]

        for (imp <- imports) {
          var filename: String = ""
          var ontmorphism: OntMorphism = IdMorph
          //println("imp = "+imp)
          imp match {
            case _ ~ fname ~ optwith =>
              filename = fname
              optwith match {
                case Some(_ ~ morph) => ontmorphism = morph
                case _ =>
              }
          }
          // println("Morph "+ontmorphism)
          // println("Filename: "  + filename)
          if (filename.endsWith(".owl") || filename.endsWith(".rdf")) {
            OWL2Ontology.fromIRIString(filename) match {
              case Some(subo) =>
                val o = ontmorphism(subo)
                tbox = (tbox ++ o.tbox).distinct
                rbox = (rbox ++ o.rbox).distinct
                abox = (abox ++ o.abox).distinct
              case None => log.error("Could not read OWL ontology " + filename)
            }
          }
          else {
            // interpret dots as path separators!
            val filenamex = (if (filename.contains(File.separator)) filename
            else currentpath + File.separator + filename.replace('.', File.separatorChar)) + ".ont"
            log.info("Import " + filenamex)
            val file = new File(filenamex)
            if (file.exists()) {
              parseAll(ontology(currentpath), new FileReader(filenamex)) match {
                case Success(subo, rem) =>
                  /*
                  println("file: " + filenamex)
                  var fout = new FileWriter(filenamex + ".res")
                  fout.write(subo.toStringLong)
                  fout.close()
                  */
                  if (rem.atEnd) {} else log.warning("remaining input: " + filenamex + " " + rem.pos)
                  val o = ontmorphism.apply(subo)
                  tbox = (tbox ++ o.tbox).distinct
                  rbox = (rbox ++ o.rbox).distinct
                  abox = (abox ++ o.abox).distinct
                case NoSuccess(msg, _) => log.error("Error parsing file " + filenamex + ":\n" + msg)
              }
            }
          }
        }
        val res = Ontology(tbox, rbox, abox)
        res match {
          case EmptyOntology =>
          case _ => // log.info("Returning\n" + res.toStringAux)
        }
        res
    }

  def pathparser: Parser[String] = opt("http://" | "file://") ~ opt("/" | "\\") ~ repsep("[^/\\\n;, ]+".r, "/" | "\\") ^^ {
    case prot ~ abs ~ paths =>
      val prefix = {
        ((prot match {
          case Some(s) => s
          case None => ""
        }): String) +
          (abs match {
            case Some(s) => File.separator
            case _ => ""
          })
      }
      //println("prefix "+prefix)
      val res = prefix + paths.mkString(File.separator)
      res.trim()
  }

  def abox: Parser[List[ABoxFormula]] = "{" ~ repsep(aboxformula, ",") ~ "}" ^^ {
    case "{" ~ ms ~ "}" => ms map (_.ground)
  }
}

case class Ontology(tbox: List[TBoxFormula], rbox: List[RBoxFormula], abox: List[ABoxFormula]) extends ABoxFormulaParser with JavaLogging {
  var reasoner: Option[PelletUpdateReasoner] = None
  var primitiveConcepts: Option[Set[Concept]] = None
  var primitiveRoles: Option[Set[Role]] = None
  var ancestor: Option[Ontology] = None
  var diff: Option[(List[ABoxFormula], List[ABoxFormula])] = None
  var indirecteffectrules = Map.empty[String, IndirectEffect]
  val pm = Builtin.prefixManager
  var conceptInds = Map.empty[Concept, List[String]]
  var allInds = List[String]()
  var counter: Int = 0
  val base: String = "r"
  var annotationMap: scala.collection.mutable.Map[OWLObject, OWLAnnotation] = scala.collection.mutable.Map.empty[OWLObject, OWLAnnotation]

  /**
   * takes an ontology as input and produces a generator for new individuals
   * needs to be regenerated when new individuals are added from outside the function
   */
  def newIndividualGenerator: String = {

    if (allInds.isEmpty) allInds = allIndividuals
    //println("all inds = "+allInds)
    while (allInds.contains(base + counter)) counter = counter + 1
    allInds = base + counter :: allInds
    base + counter
  }

  def interruptReasoner() {
    reasoner match {
      case Some(r) => r.interrupt()
      case None =>
    }
  }

  def setAncestor(o: Ontology) {
    ancestor = Some(o)
  }

  def deleteAncestor() {
    ancestor = None
  }

  def addIndirectEffectRules(rules: Map[String, DLDeclaration]) {
    for (e <- rules) {
      e._2 match {
        case effect: IndirectEffect => indirecteffectrules = indirecteffectrules updated(e._1, effect)
        case _ =>
      }
    }
  }

  /*
    Computes the difference to the last Ontology ABox as a Pair of deleted and added ABoxFormulas
   */

  def getDiff: (List[ABoxFormula], List[ABoxFormula]) = {
    diff match {
      case None => ancestor match {
        case None =>
          val res = Pair(List.empty[ABoxFormula], abox)
          diff = Some(res)
          res
        case Some(o) =>
          val res = computeDiff(o)
          diff = Some(res)
          res
      }
      case Some(res) => res
    }
  }

  def addAnnotations(annoMap: scala.collection.mutable.Map[OWLObject, OWLAnnotation]) {
    annotationMap.++=(annoMap)
  }

  /*
    Checks if the diff to the previous ontology is empty
   */
  def diffIsEmpty: Boolean = {
    getDiff ==(Nil, Nil)
  }

  def computeDiff(old: Ontology): Pair[List[ABoxFormula], List[ABoxFormula]] = {
    var added = List.empty[ABoxFormula]
    var deleted = List.empty[ABoxFormula]
    val isnegated: ABoxFormula => Boolean = {
      case ConceptAssertion(n, c) => c match {
        case NotConcept(_) => false
        case _ => false
      }
      case RoleAssertion(s, t, r) => r match {
        case NotRole(_) => true
        case _ => false
      }
      case _ => false
    }
    for (newformula: ABoxFormula <- abox diff old.abox) {
      if (isnegated(newformula)) deleted = OntologyFun.Neg(newformula) :: deleted
      else added = newformula :: added
    }
    for (deletedformula: ABoxFormula <- old.abox diff abox) {
      if (isnegated(deletedformula)) added = OntologyFun.Neg(deletedformula) :: added
      else deleted = deletedformula :: deleted
    }
    Pair(deleted, added)
  }


  override def toString = {
    /* "%TBOX\n"+
      ((tbox :\ "") ((a,b) => a+"\n"+b))+"\n\n%RBOX\n"+
      ((rbox :\ "") ((a,b) => a+"\n"+b))+"\n\n"+
      ((abox :\ "") ((a,b) => a+"\n"+b))+
      "indirect effect rules: "+(indirecteffectrules.keySet)+"\n"+
     */ getDiff match {
      case Pair(del, add) => "\nOntology: Diff:\n" +
        (del :\ "")((a, b) => "-" + a + "\n" + b) +
        (add :\ "")((a, b) => "+" + a + "\n" + b)
    }
  }

  def toStringAux: String = {
    "%TBOX\n" +
      (tbox :\ "")((a, b) => a + "\n" + b) + "\n%RBOX\n" +
      (rbox :\ "")((a, b) => a + "\n" + b) + "\n" +
      (abox :\ "")((a, b) => a + "\n" + b)
  }

  def toStringLong: String = {
    toStringAux +
      "indirect effect rules: " + indirecteffectrules.keySet + "\n" +
      "primitive concepts: " + computePrimitiveConcepts + "\n" + {
      (computePrimitiveConcepts :\ "") {
        (c, s) => {
          val res = conceptGetDefiningConceptsAndRoles(c)
          if (res.head.isEmpty) s
          else s + "  " + c + " defining roles: " + res + "\n"
        }
      }
    } +
      "primitive roles: " + computePrimitiveRoles + "\n" +
      toString +
      "Superfluous facts: " + (superflousAboxFacts :\ "\n")((phi, s) => s + "  " + phi + "\n") +
      "Ambiguous: " + !isUnambiguous
  }

  /**
   * takes a list of preconditions (and) and returns all possible substitutions
   * @todo auf pellet umbauen
   */

  def query(q: ABoxFormula): List[Substitution] = {
    //println("reasoner = "+reasoner) ;
    query(List(q))
  }

  def query(s: String): List[Substitution] = queryGround(s, groundFormulas = false)

  def queryGround(s: String, groundFormulas: Boolean): List[Substitution] = {
    parseAll(aboxformulas, s) match {
      case Success(formulas, _) =>
        query(if (groundFormulas) formulas map (_.ground) else formulas)
      case _ => log.error("Error parsing '" + s + "'"); Nil
    }
  }

  def query2(phi: ABoxFormula): List[Substitution] = {
    phi match {
      case ConceptAssertion(Const(n), concept) =>
        if (allIndividuals(concept).contains(n)) List(Id()) else Nil
      case ConceptAssertion(v@Var(n), concept) =>
        allIndividuals(concept) map (i => /(Const(i), v))
      case RoleAssertion(src, trg, role) =>
        allMembers(role).toList flatMap {
          case (s, t) => (src, trg) match {
            case (Const(x), Const(y)) if x == s && y == t => List(Id())
            case (u@Var(x), Const(y)) if y == t => List(/(Const(s), u))
            case (Const(x), v@Var(y)) if x == s => List(/(Const(t), v))
            case (u@Var(x), v@Var(y)) => List(CompSubst(/(Const(s), u), /(Const(t), v)))
            case _ => Nil
          }
        }
      case SameAsAssertion(src, trg) => (src, trg) match {
        case (u@Var(x), v@Var(y)) => allIndividuals flatMap (i => {
          getSameIndividuals(i).toList flatMap (o => {
            List(CompSubst(/(Const(i), u), /(Const(o), v)))
          })
        })
        case (Const(x), v@Var(y)) =>
          getSameIndividuals(x).toList flatMap (o => {
            List(/(Const(o), v))
          })
        case (u@Var(x), Const(y)) =>
          getSameIndividuals(y).toList flatMap (i => {
            List(/(Const(i), u))
          })
        case (Const(x), Const(y)) =>
          if (getSameIndividuals(x).contains(y)) List(Id())
          else Nil
        case _ => throw new Exception("query2.1 " + phi)
      }

      case DifferentFromAssertion(src, trg) => (src, trg) match {
        case (u@Var(x), v@Var(y)) => allIndividuals flatMap (i => {
          getDifferentIndividuals(i).toList flatMap (o => {
            List(CompSubst(/(Const(i), u), /(Const(o), v)))
          })
        })
        case (Const(x), v@Var(y)) =>
          getDifferentIndividuals(x).toList flatMap (o => {
            List(/(Const(o), v))
          })
        case (u@Var(x), Const(y)) =>
          getSameIndividuals(y).toList flatMap (i => {
            List(/(Const(i), u))
          })
        case (Const(x), Const(y)) =>
          if (getDifferentIndividuals(x).contains(y)) List(Id())
          else Nil
        case _ => throw new Exception("query2.2 " + phi)
      }
      case _ => throw new Exception("query2: " + phi)
    }
  }

  def query2(phis: List[ABoxFormula]): List[Substitution] = query2(phis, Id())

  private def query2(phis: List[ABoxFormula], subst: Substitution): List[Substitution] = {
    phis match {
      case Nil => List(subst)
      case phi :: phirest => query2(phi) flatMap {
        s => query2(s.apply(phirest), CompSubst(s, subst))
      }
    }
  }

  private def createReasoner() {
    this.synchronized {
      reasoner match {
        case None =>
          //println("Start reasoner ")
          //saveToOWLFile(new File("/tmp/ship.owl"))
          //manager.saveOntology(owlont,System.out)
          reasoner = Some(new PelletUpdateReasoner(this))
        //primitiveConcepts = Ontology2OWL.computeTopClasses(reasoner.get.owlont) map (owlclass => ConceptConstant(owlclass.getIRI.getFragment))
        //println("Primitive concepts: "+(primitiveConcepts.length))
        //println("Done create reasoner")
        /* try {
        throw new Exception("Here")
      } catch {
        case (e:Exception) => e.printStackTrace()
      }  */
        case Some(r) => //println("Reasoner already started")
      }
    }
  }

  def allConcepts: Set[Concept] =
    (tbox :\ Set.empty[Concept]) {
      (phi, s) => s ++ phi.getConcepts
    } ++
      (rbox :\ Set.empty[Concept]) {
        (phi, s) => s ++ phi.getConcepts
      } ++
      (abox :\ Set.empty[Concept]) {
        (phi, s) => s ++ phi.getConcepts
      }

  def allRoles: Set[Role] =
    (tbox :\ Set.empty[Role]) {
      (phi, s) => s ++ phi.getRoles
    } ++
      (rbox :\ Set.empty[Role]) {
        (phi, s) => s ++ phi.getRoles
      } ++
      (abox :\ Set.empty[Role]) {
        (phi, s) => s ++ phi.getRoles
      }

  private def getNamedSubconcepts(c: Concept): Set[Concept] = {
    c match {
      case ConceptConstant(n) => Set(c)
      case AndConcept(c1, c2) => getNamedSubconcepts(c1) ++ getNamedSubconcepts(c2)
      case OrConcept(c1, c2) => getNamedSubconcepts(c1) ++ getNamedSubconcepts(c2)
      case _ => Set.empty
    }
  }

  private def computePrimitiveConcepts: Set[Concept] = {
    primitiveConcepts match {
      case None =>
        var primConcepts = allConcepts
        tbox foreach {
          case ConceptInclusion(c1: ConceptConstant, c2) =>
            //println("Testing conceptinclusion "+c1+" in "+c2)
            //println("Equivalent "+isEquivalentConcept(c1,c2))
            if (isEquivalentConcept(c1, c2)) primConcepts = primConcepts - c1
          //else primConcepts = primConcepts -- getNamedSubconcepts(c2)
          case ConceptEquality(c1: ConceptConstant, c2) =>
            if (isEquivalentConcept(c1, c2)) primConcepts = primConcepts - c1
          case ConceptEquality(c1, c2: ConceptConstant) =>
            if (isEquivalentConcept(c2, c1)) primConcepts = primConcepts - c2
          case _ =>
        }
        primitiveConcepts = Some(primConcepts)
        primConcepts
      case Some(p) => p
    }
  }

  def superflousAboxFacts: List[ABoxFormula] = {
    abox filterNot {
      phi => isNotSuperfluous(phi)
    }
  }

  def isNotSuperfluous(phi: ABoxFormula): Boolean = {
    phi match {
      case ConceptAssertion(_, NotConcept(c)) => conceptIsPrimitive(c)
      case ConceptAssertion(_, c) => conceptIsPrimitive(c)
      case RoleAssertion(_, _, NotRole(r)) => roleIsPrimitive(r)
      case RoleAssertion(_, _, r) => roleIsPrimitive(r)
      case _ => true
    }
  }

  def conceptIsPrimitive(c: Concept): Boolean = {
    computePrimitiveConcepts.contains(c)
  }

  def isUnambiguous: Boolean = {
    computePrimitiveConcepts forall (concept => {
      val varx = Var("i")
      query(concept.mkAssertion(varx)) forall (subst => {
        val constant = subst(varx).asInstanceOf[Const]
        conceptGetDefiningConceptsAndRoles(concept) exists (roles => {
          roles forall {
            case concept: Concept =>
              val res = !query(List(concept.mkAssertion(constant))).isEmpty
              //println("Checking "+concept(constant)+" = "+res)
              res
            case role: Role =>
              val vary = Var("o")
              val res = !query(List(role.roleAssert(constant, vary))).isEmpty
              //println("Checking "+role(constant, vary)+" = "+res)
              res
          }
        })
      })
    })
  }

  def computePrimitiveRoles: Set[Role] = primitiveRoles match {
    case None =>
      var primRoles = allRoles
      rbox foreach {
        case RoleInclusion(roles, role) => primRoles = primRoles - role
        case RoleEquality(roles, role) => primRoles = primRoles - role
        case _ =>
      }
      primitiveRoles = Some(primRoles)
      primRoles
    case Some(r) => r
  }

  def roleIsPrimitive(r: Role): Boolean = {
    r match {
      case r: DataRole => true
      case _ => computePrimitiveRoles.contains(r)
    }
  }

  private def conceptGetDefiningConceptsAndRoles(c: Concept): Set[Set[ConceptOrRole]] = {
    if (conceptIsPrimitive(c)) {
      createReasoner()
      val supclasses: Set[Concept] = OWLQuery.getSuperClasses(c, pm, reasoner.get.reasoner) + c

      {
        (tbox :\ Set(Set.empty[ConceptOrRole])) {
          (phi, currents) => phi match {
            case ConceptInclusion(sub, sup) =>
              if (supclasses.contains(sub)) {
                (currents :\ Set.empty[Set[ConceptOrRole]])((current, result) => {
                  val res = getPrimitiveConceptsAndRoles(sup, current)
                  //println("Getdefining roles from "+sup+" = "+res)
                  result ++ res
                })
              }
              else currents
            case _ => currents
          }
        }
      }
    } else Set.empty[Set[ConceptOrRole]]
  }

  private def getDefiningPrimitiveRole(r: Role): Role = {
    val res = if (roleIsPrimitive(r)) r
    else {
      {
        rbox collectFirst {
          case RoleInclusion(roles, role) if r == role => getDefiningPrimitiveRole(roles.head)
          case RoleEquality(roles, role) if r == role => getDefiningPrimitiveRole(roles.head)
        }
      }.getOrElse(r)
    }
    //println(r+" defining prim role "+res)
    res
  }

  private def getPrimitiveConceptsAndRoles(c: Concept, current: Set[ConceptOrRole]): Set[Set[ConceptOrRole]] = {
    c match {
      case AndConcept(c1, c2) =>
        (getPrimitiveConceptsAndRoles(c1, current) :\ Set(Set.empty[ConceptOrRole]))((newcurrent, result) => {
          result ++ getPrimitiveConceptsAndRoles(c2, newcurrent)
        })
      case OrConcept(c1, c2) => getPrimitiveConceptsAndRoles(c1, current) ++ getPrimitiveConceptsAndRoles(c2, current)
      case ExistsRole(r, c1) => Set(current + getDefiningPrimitiveRole(r))
      case ForallRole(r, c1) => Set(current + getDefiningPrimitiveRole(r))
      case _ => Set(Set.empty[ConceptOrRole])

    }
  }

  def getReasoner: OWLReasoner = {
    reasoner match {
      case None =>
        createReasoner()

        getReasoner
      case Some(r) => r.reasoner
    }
  }

  def getManager: OWLOntologyManager = {
    reasoner match {
      case None =>
        createReasoner()
        getManager
      case Some(r) => r.manager
    }
  }


  def isConsistent: Boolean = {
    reasoner match {
      case None =>
        createReasoner()
        isConsistent
      case Some(r) => r.reasoner.isConsistent
    }
  }

  def isUnsatisfiable(ax: ABoxFormula): Boolean = {
    isUnsatisfiable(List(ax))
  }

  def allindividualsdeclared(axs: List[ABoxFormula]): Boolean =
    axs forall {
      case ConceptAssertion(x, _) => x match {
        case Const(n) => allIndividuals.contains(n)
        case _ => true
      }
      case RoleAssertion(u, v, _) =>
        (u match {
          case Const(s) => allIndividuals.contains(s)
          case _ => true
        }) ||
          (v match {
            case Const(t) => allIndividuals.contains(t)
            case _ => true
          })
      case DifferentFromAssertion(u, v) =>
        (u match {
          case Const(s) => allIndividuals.contains(s)
          case _ => true
        }) ||
          (v match {
            case Const(t) => allIndividuals.contains(t)
            case _ => true
          })
      case SameAsAssertion(u, v) =>
        (u match {
          case Const(s) => allIndividuals.contains(s)
          case _ => true
        }) ||
          (v match {
            case Const(t) => allIndividuals.contains(t)
            case _ => true
          })
      case _ => false
    }

  val unsatcache = scala.collection.mutable.HashMap.empty[List[ABoxFormula], Boolean]

  def isUnsatisfiable(axs: List[ABoxFormula]): Boolean = {
    val key = axs.sortWith({
      (a, b) => a.toString < b.toString
    })
    unsatcache.get(key) match {
      case Some(res) => res
      case None =>
        this.synchronized {
          var result = abox exists (phi => containsInconsistentFormula(phi, axs))

          if (!result && allindividualsdeclared(axs)) {
            val newabox = axs ++ abox

            // Include old abox facts if consistent

            val newo = Ontology(tbox, rbox, newabox)

            //println("Old reasoner "+reasoner)
            // Adjusting the old reasoner (if ti existed) to this ontology
            reasoner match {
              case None => createReasoner()
              case _ =>
            }
            /* result = !newo.isConsistent  */
            reasoner match {
              case Some(r) =>
                val additionnalupdates = r.update(newo)
                //println("Additionnalupdate "+additionnalupdates)
                result = additionnalupdates != Pair(Nil, Nil)
                r.update(this)
              case v => v
            }
          } else {
            result = true
          }
          /*
          if (result) println("Unsat "+axs)
          else println ("sat "+axs)
          */
          unsatcache.put(key, result)
          result
        }
    }
  }

  def getRoleSignature(r: Role): Pair[Concept, Concept] = {
    r match {
      case NamedRole(n) =>
        val optres = rbox find {
          case RoleDeclaration(r1, dom, ran) => r == r1
          case _ => false
        }
        optres match {
          case Some(RoleDeclaration(r1, dom, ran)) => Pair(dom, ran)
          case _ => Pair(UniversalConcept, UniversalConcept)
        }
      case AndRole(r1, r2) => getRoleSignature(r1) match {
        case (dom1, ran1) => getRoleSignature(r2) match {
          case (dom2, ran2) => (AndConcept(dom1, dom2), AndConcept(ran1, ran2))
        }
      }
      case OrRole(r1, r2) => getRoleSignature(r1) match {
        case (dom1, ran1) => getRoleSignature(r2) match {
          case (dom2, ran2) => (OrConcept(dom1, dom2), OrConcept(ran1, ran2))
        }
      }
      case InverseRole(r1) => getRoleSignature(r1).swap
      case NominalRole(src, trg) => (Nominal(Const(src)), Nominal(Const(trg)))
      case _ => (UniversalConcept, UniversalConcept)
    }
  }

  def roleIsFunctional(r: Role): Boolean = {
    val res = rbox exists {
      case RoleFun(r1) => r == r1
      case _ => false
    }
    //println(r.toString+" is functional "+res)
    res
  }

  def roleIsInverseFunctional(r: Role): Boolean = {
    rbox exists {
      case RoleInvFun(r1) => r == r1
      case _ => false
    }
  }

  def saveToOWLFile(f: File) {
    val manager: OWLOntologyManager = OWLManager.createOWLOntologyManager()
    val owlont = Ontology2OWL.convert(this, Builtin.prefixManager, manager)
    try {
      val out = new FileOutputStream(f)
      val ren = new OWLXMLRenderer()
      ren.render(owlont, out)
      //manager.saveOntology(owlont, out)
      out.close()
    } catch {
      case ex: Exception => log.error(ex, "saveToOWLFile")
    }
  }

  def explainInferences(formulas: List[ABoxFormula], subst: Substitution): String = {
    createReasoner()
    val expGen: PelletExplanation = new PelletExplanation(reasoner.get.reasoner)
    val sw = new StringWriter()
    formulas foreach {
      formula =>
        val sformula = subst(formula)
        val owlaxiom = Ontology2OWL.aboxformula2OWL(sformula, reasoner.get.owlont, pm, reasoner.get.manager)
        sw.append(sformula.toString + " (subst = " + subst + ") follows from " +
          (JavaConversions.asScalaSet(expGen.getEntailmentExplanation(owlaxiom)) :\ "")
            ((form, str) => str + "\n " + form) + "\n")
    }
    sw.toString
  }

  val querycache = scala.collection.mutable.HashMap.empty[List[ABoxFormula], List[Substitution]]


  def query(qs: List[ABoxFormula]): List[Substitution] = {
    if (this == EmptyOntology) List.empty
    else {
      // query2(qs)
      val key = qs.sortWith({
        (a, b) => a.toString < b.toString
      })
      querycache.get(key) match {
        case Some(res) => res
        case None =>
          val result = {
            if (qs.isEmpty) List(Id())
            else this.synchronized {
              //println("reasoner ="+reasoner)
              reasoner match {
                case None =>
                  createReasoner()
                  query(qs)
                case Some(r) =>

                  //r.reasoner.prepareReasoner()
                  //r.reasoner.getKB.classify()
                  //r.reasoner.getKB.realize()

                  val qs1 = if (isBooleanAbox) qs map booleanAssertion2Assertion else qs
                  if (allindividualsdeclared(qs1)) {
                    val q: de.derivo.sparqldlapi.Query = OWLQuery.aBoxFormulas2Query(qs1, Builtin.prefixManager)
                    try {
                      val substs = OWLQuery.queryResult2Substitution(r.engine.execute(q), q)
                      // Serge@ALL: Uncomment the follwing line if you want to see an explanation set for the queries
                      // substs foreach (s => println(explainInferences(qs1,s)))
                      substs
                    }
                    catch {
                      case ex: Exception =>
                        createReasoner()
                        log.info("query exception: " + ex + "\nquery: " + q)
                        List.empty[Substitution]
                    }
                  }
                  else List.empty[Substitution]
              }
            }
          }
          log.info("query result " + qs + ": " + result)
          // log.info("query2 result " + qs + ": " + query2(qs))

          querycache.put(key, result)
          result
      }
    }
  }

  def disjointWith(c1: Concept, c2: Concept): Boolean = this.synchronized {
    createReasoner()
    val q: de.derivo.sparqldlapi.Query =
      OWLQuery.createDisjointnessQuery(Builtin.prefixManager, c1, c2)
    try {
      !OWLQuery.queryResult2Substitution(reasoner.get.engine.execute(q), q).isEmpty
    }
    catch {
      case ex: Exception => log.error(ex, "disjointWith")
        createReasoner()
        false
    }
  }


  /* TODO Finish implementation
  def getRolePairs(role:Role):List[ABoxFormula] = this.synchronized {

    reasoner match {
      case None =>
        createReasoner()
        getRolePairs(role)
      case Some(r) => OWLQuery.RolePairs(concept, Builtin.prefixManager, r)

    }
  }
  */


  /**
   *
   */
  def allIndividuals(concept: Concept): List[String] =
    conceptInds get concept match {
      case Some(l) => l
      case None => this.synchronized {
        reasoner match {
          case None =>
            createReasoner()
            allIndividuals(concept)
          case Some(r) =>
            val res = OWLQuery.getIndividuals(concept,
              Builtin.prefixManager,
              r.reasoner)
            //println("All individuals of "+concept+" = "+res.toList)
            conceptInds = conceptInds updated(concept, res)
            res
        }
      }
    }

  def allIndividuals: List[String] = this.synchronized {
    reasoner match {
      case None =>
        createReasoner()
        allIndividuals
      case Some(r) =>
        var result: List[String] = List.empty[String]
        val it = r.reasoner.getKB.getIndividuals.iterator()
        while (it.hasNext) {
          var i = it.next().getName
          i = i.substring(i.indexOf("#") + 1)
          result = i :: result
        }
        //println("all individuals "+result)
        result
    }
  }

  def isEquivalentConcept(c: ConceptConstant, cexpr: Concept): Boolean = this.synchronized {

    reasoner match {
      case None => createReasoner(); isEquivalentConcept(c, cexpr)
      case Some(r) => OWLQuery.isEquivalentConcept(c, cexpr, pm, r.reasoner)
    }
  }

  def isLowestLevelConcept(c: ConceptConstant): Boolean = this.synchronized {

    reasoner match {
      case None => createReasoner(); isLowestLevelConcept(c)
      case Some(r) => OWLQuery.isLowestLevelClass(c, pm, r.reasoner)
    }
  }

  def allMembers(role: Role): Set[(String, String)] = this.synchronized {
    reasoner match {
      case None =>
        createReasoner()
        allMembers(role)
      case Some(r) =>
        val res = OWLQuery.getRoleMembers(role,
          Builtin.prefixManager,
          r.reasoner)
        // log.info("All members of " + role + " = " + res.toList)
        res
    }
  }

  def getDifferentIndividuals(ind: String): Set[String] = this.synchronized {
    reasoner match {
      case None =>
        createReasoner()
        getDifferentIndividuals(ind)
      case Some(r) =>
        val res = OWLQuery.getDifferentIndividuals(ind,
          Builtin.prefixManager,
          r.reasoner)
        log.info("All different individuals from " + ind + " = " + res.toList)
        res
    }
  }

  def getSameIndividuals(ind: String): Set[String] = this.synchronized {
    reasoner match {
      case None =>
        createReasoner()
        getDifferentIndividuals(ind)
      case Some(r) =>
        val res = OWLQuery.getSameIndividuals(ind,
          Builtin.prefixManager,
          r.reasoner)
        log.info("All same individuals from " + ind + " = " + res.toList)
        res
    }
  }

  def mycontains(phi: ABoxFormula, phis: List[ABoxFormula]): Boolean = {
    val res = phis exists (p => p == phi)
    //if (res) println("Contains "+phi) else println("Not contains "+phi)
    res
  }


  def updatesimple(update: List[ABoxFormula], deleteIndividuals: List[VarOrConstant]): Option[Ontology] = {
    //println("Update: "+update)

    /* Ontology(tbox,rbox,
       ((abox filter (phi => !(mycontains(OntologyFun.Neg(phi),update))))
       ++
       (update flatMap (phi => if (abox contains phi) Nil else List(phi)))).distinct
   )   */
    val newont = updateIncremental(update, deleteIndividuals)
    newont
    // booleanUpdate(update)
  }

  def containsInconsistentFormula(phi: ABoxFormula, phis: List[ABoxFormula]): Boolean = {
    phi match {
      case v@RoleAssertion(s, t, r) =>
        mycontains(OntologyFun.Neg(phi), phis) ||
          roleIsFunctional(r) && (phis exists {
            case RoleAssertion(s1, t1, r1) => r == r1 && s == s1 && t != t1 && {
              log.info("Removing " + v)
              true
            }
            case _ => false
          }) ||
          roleIsInverseFunctional(r) && (phis exists {
            case RoleAssertion(s1, t1, r1) => r == r1 && s != s1 && t == t1 && {
              log.info("Removing " + v)
              true
            }
            case _ => false
          })
      case x@ConceptAssertion(v, c) =>
        mycontains(OntologyFun.Neg(phi), phis) ||
          (phis exists {
            case ConceptAssertion(v1, c1) => v == v1 && disjointWith(c, c1) && {
              log.info("Removing " + x)
              true
            }
            case _ => false
          })
      case _ => false
    }
  }

  def updateIncremental(update: List[ABoxFormula], deleteIndividuals: List[VarOrConstant]): Option[Ontology] = this.synchronized {
    val check = update filterNot (phi => isNotSuperfluous(phi))
    if (!check.isEmpty) {
      log.warning("Following updates are superfluous: " + check)
    }
    var purifyoldabox = abox filterNot (phi => containsInconsistentFormula(phi, update))

    //var allinds = allIndividuals

    for (i: VarOrConstant <- deleteIndividuals) {
      purifyoldabox = removeIndividual(i, purifyoldabox)
      //allinds = allinds - i.name
    }

    var newabox = update filter {
      case RoleAssertion(s, t, NotRole(r)) => false
      case ConceptAssertion(v, NotConcept(c)) => false
      case _ => true
    }

    // Applying exhaustively indirect effect rules on newabox

    newabox = computeIndirectEffects(newabox)
    /*allinds = (allinds ++ (newabox flatMap (p =>
      p match {
        case ConceptAssertion(Const(x),_) => List(x)
        case RoleAssertion(Const(u),Const(v),_) =>
          List(u,v)
      }))).distinct    */

    // Adding the old purified abox entries

    newabox = (newabox ++ purifyoldabox).distinct

    // Include old abox facts if consistent

    var newo = Ontology(tbox, rbox, newabox)

    //println("Old reasoner "+reasoner)
    // Adjusting the old reasoner (if ti existed) to this ontology
    reasoner match {
      case None => createReasoner()
      case _ =>
    }
    reasoner match {
      case Some(r) =>
        val additionnalupdates = r.update(newo)
        newabox = newabox filterNot (ax => additionnalupdates._1.contains(ax))
        newabox = newabox ++ additionnalupdates._2
        newo = Ontology(tbox, rbox, newabox)
        newo.reasoner = Some(r)
        newo.annotationMap = annotationMap
        reasoner = None
      case v => v
    }

    // Adding the indirect rules

    newo.addIndirectEffectRules(indirecteffectrules)

    // Setting the ancestor relation

    newo.setAncestor(this)
    newo.getDiff
    newo.deleteAncestor()

    // Returning the new ontology

    Some(newo)
  }


  def computeIndirectEffects(abox: List[ABoxFormula]): List[ABoxFormula] = {
    var newabox = abox
    val res = indirecteffectrules find (e => {
      val rule = e._2
      //println("Testing rule "+e._1)
      rule.apply(this, abox) match {
        case Some(effects) => newabox = effects; true
        case None => false
      }
    })
    res match {
      case Some((n, r)) =>
        //println("Applied indirect effect rule "+n)
        computeIndirectEffects(newabox)
      case None => abox
    }
  }

  def isBooleanAbox: Boolean = {
    abox exists (phi => phi.isInstanceOf[Disjunction])
  }

  def update(pos: List[ABoxFormula], neg: List[ABoxFormula], deleteIndividuals: List[VarOrConstant]): Option[Ontology] = {
    updatesimple(pos ++ (neg map (phi => OntologyFun.Neg(phi))), deleteIndividuals)
  }

  def booleanUpdate(update: List[ABoxFormula]): Ontology = {
    // Allowing only primitive updates
    assert(update forall (phi => phi.isPrimitive))
    val newabox = (abox flatMap (phi => booleanUpdate(phi, update))) ++ update
    Ontology(tbox, rbox, newabox)
  }

  def booleanUpdate(phi: ABoxFormula, update: List[ABoxFormula]): List[ABoxFormula] = {
    phi match {
      case ConceptAssertion(n, c) =>
        if (update exists {
          case ConceptAssertion(n1, c1) => n1 == n && c1.isNegationOf(c)
          case _ => false
        }) List.empty[ABoxFormula]
        else {
          val up = ConceptAssertion(n, booleanUpdate(c, update))
          if (up == phi) List[ABoxFormula](phi) else List[ABoxFormula](Disjunction(phi, up))
        }
      case RoleAssertion(s, t, r) =>
        if (update exists {
          case RoleAssertion(s1, t1, r1) => if (s1 == s && t1 == t && r1.isNegationOf(r)) true else false
          case _ => false
        }) List.empty[ABoxFormula]
        else {
          val up = RoleAssertion(s, t, booleanUpdate(r, update))
          if (up == phi) List[ABoxFormula](phi) else List[ABoxFormula](Disjunction(phi, up))
        }
      case Disjunction(f, g) =>
        val up1 = booleanUpdate(f, update)
        val up2 = booleanUpdate(g, update)
        up1 match {
          case f1 :: _ => up2 match {
            case g1 :: _ => List[ABoxFormula](Disjunction(f1, g1))
            case Nil => List[ABoxFormula](f1)
          }
          case Nil => up2 match {
            case g1 :: _ => List[ABoxFormula](g1)
            case Nil => List.empty[ABoxFormula]
          }
        }
      case _ => throw new Exception("booleanUpdate: " + phi)
    }
  }

  def booleanUpdate(c: Concept, update: List[ABoxFormula]): Concept = c match {
    case NotConcept(sub) => booleanUpdate(sub, update).Negate
    case AndConcept(sub1, sub2) => AndConcept(booleanUpdate(sub1, update), booleanUpdate(sub2, update))
    case OrConcept(sub1, sub2) => OrConcept(booleanUpdate(sub1, update), booleanUpdate(sub2, update))
    case ForallRole(r, sub) => ForallRole(booleanUpdate(r, update), booleanUpdate(sub, update))
    case ExistsRole(r, sub) => ExistsRole(booleanUpdate(r, update), booleanUpdate(sub, update))
    case _ =>
      var newconcept = c
      val negatedNominals = update flatMap {
        case ConceptAssertion(n1, c1) => if (c1.isNegationOf(c)) List[VarOrConstant](n1) else Nil
        case _ => Nil
      }
      val unnegatedNominals = update flatMap {
        case ConceptAssertion(n1, c1) => if (c1 == c) List[VarOrConstant](n1) else Nil
        case _ => Nil
      }
      for (n: VarOrConstant <- unnegatedNominals)
        newconcept = AndConcept(newconcept, NotConcept(Nominal(n)))
      for (n: VarOrConstant <- negatedNominals)
        newconcept = OrConcept(newconcept, Nominal(n))
      newconcept
  }

  def booleanUpdate(r: Role, update: List[ABoxFormula]): Role = r match {
    case NotRole(sub) => NotRole(booleanUpdate(sub, update))
    case InverseRole(sub) => InverseRole(booleanUpdate(sub, update))
    case AndRole(sub1, sub2) => AndRole(booleanUpdate(sub1, update), booleanUpdate(sub2, update))
    case OrRole(sub1, sub2) => OrRole(booleanUpdate(sub1, update), booleanUpdate(sub2, update))
    case _ =>
      var newrole = r
      val negatedNominals = update flatMap {
        case RoleAssertion(s1, t1, r1) => if (r1.isNegationOf(r)) List[Pair[VarOrConstant, VarOrConstant]](Pair(s1, t1)) else Nil
        case _ => Nil
      }
      val unnegatedNominals = update flatMap {
        case RoleAssertion(s1, t1, r1) => if (r1 == r) List[Pair[VarOrConstant, VarOrConstant]](Pair(s1, t1)) else Nil
        case _ => Nil
      }
      for ((n1, n2) <- unnegatedNominals)
        newrole = AndRole(newrole, NotRole(NominalRole(n1.toString, n2.toString)))
      for ((n1, n2) <- negatedNominals)
        newrole = OrRole(newrole, NominalRole(n1.toString, n2.toString))
      newrole
  }

  def toClassicAboxWithAt: Ontology = {
    val newabox = abox map booleanAssertion2AssertionWithAt
    Ontology(tbox, rbox, newabox)
  }

  def booleanAssertion2AssertionWithAt(phi: ABoxFormula): ABoxFormula = {
    phi match {
      case Disjunction(_, _) => ConceptAssertion(universalIndividual, assertion2conceptWithAt(phi))
      case _ => ConceptAssertion(universalIndividual, assertion2conceptWithAt(phi))
    }
  }

  def assertion2conceptWithAt(phi: ABoxFormula): Concept = {
    phi match {
      case ConceptAssertion(n, c) => At(n, c)
      case RoleAssertion(s, t, r) => At(s, ExistsRole(r, Nominal(t)))
      case Disjunction(f, g) => OrConcept(assertion2conceptWithAt(f), assertion2conceptWithAt(g))
      case _ => throw new Exception("assertion2conceptWithAt: " + phi)
    }
  }

  val universalIndividual = Const("dummy")
  val univRole = NamedRole("R")


  def toClassicAbox: Ontology = {
    val newabox = abox map booleanAssertion2Assertion
    val ont = Ontology(tbox, rbox, newabox)
    log.info("ClassicAbox: " + newabox)
    ont
  }

  def booleanAssertion2Assertion(phi: ABoxFormula): ABoxFormula = {
    phi match {
      case Disjunction(_, _) => ConceptAssertion(universalIndividual, assertion2concept(phi))
      case _ => ConceptAssertion(universalIndividual, assertion2concept(phi))
    }
  }

  def assertion2concept(phi: ABoxFormula): Concept = {
    removeAtConcepts(assertion2conceptWithAt(phi))
  }

  def removeAtConcepts(c: Concept): Concept = {
    findAtConcept(c) match {
      case None => c
      case Some(at@At(n, sub)) =>
        val newc =
          OrConcept(
            AndConcept(Nominal(n), AndConcept(ExistsRole(univRole, sub), replaceConcept(c, at, UniversalConcept))),
            AndConcept(Nominal(n), AndConcept(ExistsRole(univRole, sub.Negate), replaceConcept(c, at, UniversalConcept.Negate)))
          )
        removeAtConcepts(newc)
    }
  }

  def findAtConcept(c: Concept): Option[At] = {
    c match {
      case v@At(_, _) => Some(v)
      case NotConcept(sub) => findAtConcept(sub)
      case OrConcept(c1, c2) => findAtConcept(c1) orElse findAtConcept(c2)
      case AndConcept(c1, c2) => findAtConcept(c1) orElse findAtConcept(c2)
      case ForallRole(r, sub) => findAtConcept(sub)
      case ExistsRole(r, sub) => findAtConcept(sub)
      case MinRestriction(_, _, sub) => findAtConcept(sub)
      case MaxRestriction(_, _, sub) => findAtConcept(sub)
      case _ => None // Nominal,Constant,Var, UniversalConcept
    }
  }


  def replaceConcept(c: Concept, oldc: Concept, newc: Concept): Concept = {
    val recurse = (sub: Concept) => replaceConcept(sub, oldc, newc)
    if (c == oldc) newc
    else
      c match {
        case At(n, sub) => At(n, recurse(sub))
        case NotConcept(sub) => recurse(sub).Negate
        case OrConcept(c1, c2) => OrConcept(recurse(c1), recurse(c2))
        case AndConcept(c1, c2) => AndConcept(recurse(c1), recurse(c2))
        case ForallRole(r, sub) => ForallRole(r, recurse(sub))
        case ExistsRole(r, sub) => ExistsRole(r, recurse(sub))
        case MinRestriction(r, n, sub) => MinRestriction(r, n, recurse(sub))
        case MaxRestriction(r, n, sub) => MaxRestriction(r, n, recurse(sub))
        case _ => c // Nominal,Constant,Var, UniversalConcept
      }
  }

  def removeIndividual(i: VarOrConstant, phis: List[ABoxFormula]): List[ABoxFormula] = {
    phis flatMap {
      case ConceptAssertion(n, c) => if (n == i) List.empty[ABoxFormula]
      else {
        List[ABoxFormula](ConceptAssertion(n, removeIndividual(i, c)))
      }
      case RoleAssertion(s, t, r) => if (s == i || t == i) List.empty[ABoxFormula]
      else {
        List[ABoxFormula](RoleAssertion(s, t, removeIndividual(i, r)))
      }
      case phi => throw new Exception("removeIndividual: " + phi)
    }
  }

  def removeIndividual(i: VarOrConstant, c: Concept): Concept = {
    var newc = c
    for (x: ConceptOrRole <- findIndividual(i, newc)) {
      x match {
        case oldc: Concept => newc = replaceConcept(newc, oldc, UniversalConcept.Negate)
        case oldrole: Role => newc = replaceRole(newc, oldrole, UniversalRole.Negate)
      }
    }
    newc
  }

  def removeIndividual(i: VarOrConstant, r: Role): Role = {
    var newr = r
    for (x: ConceptOrRole <- findIndividual(i, newr)) {
      x match {
        case oldrole: Role => newr = replaceRole(newr, oldrole, UniversalRole.Negate)
        case _ =>
      }
    }
    newr
  }

  def findIndividual(i: VarOrConstant, c: Concept): List[ConceptOrRole] = {
    c match {
      case At(n, sub) => if (n == i) c :: findIndividual(i, sub) else findIndividual(i, sub)
      case NotConcept(sub) => findIndividual(i, sub)
      case OrConcept(c1, c2) => findIndividual(i, c1) ++ findIndividual(i, c2)
      case AndConcept(c1, c2) => findIndividual(i, c1) ++ findIndividual(i, c2)
      case ForallRole(r, sub) => findIndividual(i, r) ++ findIndividual(i, sub)
      case ExistsRole(r, sub) => findIndividual(i, r) ++ findIndividual(i, sub)
      case MinRestriction(r, n, sub) => findIndividual(i, r) ++ findIndividual(i, sub)
      case MaxRestriction(r, n, sub) => findIndividual(i, r) ++ findIndividual(i, sub)
      case Nominal(n) => if (n == i) List[ConceptOrRole](c) else List.empty[ConceptOrRole]
      case _ => List.empty[ConceptOrRole] // Constant,Var, UniversalConcept
    }
  }

  def findIndividual(i: VarOrConstant, r: Role): List[ConceptOrRole] = {
    r match {
      case NotRole(sub) => findIndividual(i, sub)
      case OrRole(c1, c2) => findIndividual(i, c1) ++ findIndividual(i, c2)
      case AndRole(c1, c2) => findIndividual(i, c1) ++ findIndividual(i, c2)
      case InverseRole(sub) => findIndividual(i, sub)
      case NominalRole(s, t) => if (i.isInstanceOf[Const] && (s == i.name || t == i.name)) List[ConceptOrRole](r) else List.empty[ConceptOrRole]
      case _ => List.empty[ConceptOrRole] // Constant,Var, UniversalConcept
    }
  }


  def replaceRole(c: Concept, oldrole: Role, newrole: Role): Concept = {
    val recurse = (sub: Concept) => replaceRole(sub, oldrole, newrole)
    c match {
      case At(n, sub) => At(n, recurse(sub))
      case NotConcept(sub) => recurse(sub).Negate
      case OrConcept(c1, c2) => OrConcept(recurse(c1), recurse(c2))
      case AndConcept(c1, c2) => AndConcept(recurse(c1), recurse(c2))
      case ForallRole(r, sub) => ForallRole(replaceRole(r, oldrole, newrole), recurse(sub))
      case ExistsRole(r, sub) => ExistsRole(replaceRole(r, oldrole, newrole), recurse(sub))
      case MinRestriction(r, n, sub) => MinRestriction(replaceRole(r, oldrole, newrole), n, recurse(sub))
      case MaxRestriction(r, n, sub) => MaxRestriction(replaceRole(r, oldrole, newrole), n, recurse(sub))
      case _ => c // Nominal,Constant,Var, UniversalConcept
    }
  }

  def replaceRole(r: Role, oldrole: Role, newrole: Role): Role = {
    val recurse = (sub: Role) => replaceRole(sub, oldrole, newrole)
    if (r == oldrole) newrole
    else
      r match {
        case NotRole(sub) => NotRole(recurse(sub))
        case OrRole(sub1, sub2) => OrRole(recurse(sub1), recurse(sub2))
        case AndRole(sub1, sub2) => AndRole(recurse(sub1), recurse(sub2))
        case InverseRole(sub) => InverseRole(recurse(sub))
        case _ => r
      }
  }

  def extendBy(exto: Ontology): Ontology = {
    Ontology((tbox ++ exto.tbox).distinct,
      (rbox ++ exto.rbox).distinct,
      (abox ++ exto.abox).distinct)
  }
}


object EmptyOntology extends Ontology(Nil, Nil, Nil)

object OntologyFun {

  def Neg(phi: ABoxFormula): ABoxFormula = phi match {
    case ConceptAssertion(n, c) => ConceptAssertion(n, NotConcept(c))
    case RoleAssertion(n, m, r) => RoleAssertion(n, m, NotRole(r))
    case DifferentFromAssertion(u, v) => SameAsAssertion(u, v)
    case SameAsAssertion(u, v) => DifferentFromAssertion(u, v)
    case _ => throw new Exception("OntologyFun.Neg: " + phi)
  }


  /* def C(n:String,c:String):ConceptAssertion = {
    ConceptAssertion(Const(c),ConceptConstant(n))
  }
  def CV(n:String,c:String):ConceptAssertion = {
    ConceptAssertion(Var(c),ConceptConstant(n))
  }
  def C(n:String):ConceptConstant =  ConceptConstant(n)

  def R(r:String,n:String,m:String):RoleAssertion = RoleAssertion(Const(n),Const(m),NamedRole(r))
  def RV(r:String,n:String,m:String):RoleAssertion = RoleAssertion(Var(n),Var(m),NamedRole(r))
  def R(r:String):NamedRole = NamedRole(r)

  def personC = ConceptConstant("Person")
  def wheelchairC = ConceptConstant("WheelChair")
  def AtR = NamedRole("At")
  def TravelRequestR = NamedRole("TravelRequest")
  def AssignedR = NamedRole("Assigned")
  def abox =  List(
    ConceptAssertion(Const("adam"),personC),
    ConceptAssertion(Const("eve"),personC),
    ConceptAssertion(Const("Rolland"),wheelchairC),
    C("Location","kitchen"),
    C("Location","bedroom"),
    RoleAssertion(Const("adam"),Const("kitchen"),AtR),
    RoleAssertion(Const("adam"),Const("bedroom"),TravelRequestR)
  )
  def o = Ontology(Nil,Nil,abox)

  def phi1 = FOLTL.G(All(Var("x"),personC,FOLTL.F[ConceptConstant,ABoxFormula](Atom(RoleAssertion(Var("x"),Const("kitchen"),AtR)))))

  def phi2 = FOLTL.G(All(Var("x"),personC,
                FOLTL.Impl(Atom(RoleAssertion(Var("x"),Const("bedroom"),TravelRequestR)),
                           FOLTL.F(FOLTL.Ex(Var("y"),wheelchairC,Atom(RoleAssertion(Var("y"),Var("x"),AssignedR)))))))

  def phi3 = FOLTL.G(All(Var("x"),personC,All(Var("l1"),C("Location"),
                FOLTL.Impl(Atom(RV("TravelRequest","x","l1")),U(Atom(RV("TravelRequest","x","l1")),Atom(RV("At","x","l1"))))))  )

*/
}

/*
var personC = ConceptConstant("Person")
var wheelchairC = ConceptConstant("WheelChair")
var AtR = NamedRole("At")
var abox =  List(ConceptAssertion(Const("adam"),personC),ConceptAssertion(Const("eve"),personC),ConceptAssertion(Const("Rolland"),wheelchairC),RoleAssertion(Const("adam"),Const("kitchen"),AtR))
var o = Ontology(Nil,Nil,abox)

  var phi1 = FOLTL.G(All("x",personC,FOLTL.F[ConceptConstant,ABoxFormula](Atom(RoleAssertion(Var("x"),Const("kitchen"),AtR)))))
*/

